import re

MAX_BETA_STR = f"{0xffffffff}.{0xffff}.{0xffff}"


class OPCLIVersion:
    def __init__(self, version: str, skip_beta=False):
        if isinstance(version, int):
            version = str(version)
        version, beta_num = self._parse_beta(version)
        version_tuple = version.split(".")
        if not skip_beta:
            self._beta_num = beta_num
        else:
            self._beta_num = -1
        parts = []
        for part in version_tuple:
            parts.append(int(part, 0))
        self._parts = parts

    @property
    def beta_ver(self):
        # if this is not a beta, set beta_ver to an absurdly high value
        # so it beats any actual beta version
        # this will make lt/gt/eq comparison logic simpler
        beta_ver = None
        if self._beta_num is None:
            beta_ver = OPCLIVersion(MAX_BETA_STR, skip_beta=True)
        elif self._beta_num == -1:
            pass
        else:
            beta_ver = OPCLIVersion(self._beta_num, skip_beta=True)
        return beta_ver

    @property
    def is_beta(self) -> bool:
        _is_beta = self._beta_num is not None and self._beta_num >= 0
        return _is_beta

    def _parse_beta(self, version_string):
        regex = r".*(-beta.*)$"
        beta_num = None
        match = re.match(regex, version_string)
        if match:
            # get "-beta.01"
            beta_string = match.groups()[0]
            # strip -beta.01 from the end of version_string, leaving "2.18.0"
            version_string = re.sub(beta_string, '', version_string)
            # extract '01' from '-beta.01'
            beta_num = beta_string.split(".")[1]
            # convert beta num to an int
            beta_num = int(beta_num)
        return version_string, beta_num

    def _normalize(self, other):
        parts_self = list(self._parts)
        parts_other = list(other._parts)

        # be robust about version strings of differing lenghts
        # e.g., 11.3.1 vs 11.3.1.1
        diff_len = len(parts_self) - len(parts_other)

        # if the length difference was negative, _parts_other is longer,
        # and we need to exitend _parts self
        while diff_len < 0:
            parts_self.append(0)
            diff_len += 1
        # if diff was positive, we need to extend _parts other
        while diff_len > 0:
            parts_other.append(0)
            diff_len -= 1

        # appending the OPCLIVersion object for the beta version number
        # will allow it to be transparently compared just like all the other version parts
        # then the lt/gt/eq logic doesn't have to change
        beta_ver = self.beta_ver
        if beta_ver is not None:
            parts_self.append(beta_ver)

        beta_ver = other.beta_ver
        if beta_ver is not None:
            parts_other.append(beta_ver)

        # now 11.3.1 vs 11.3.1.1 becomes
        # 11.3.1.0 vs 11.3.1.1, and can be compared 1 to 1
        return (parts_self, parts_other)

    def __str__(self):
        beta_part = None
        if self.is_beta:
            beta_part = f"-beta.{self._beta_num:02d}"
        _str = ".".join([str(i) for i in self._parts])
        if beta_part:
            _str += beta_part
        return _str

    def __eq__(self, other):
        equal = id(self) == id(other)

        if not equal:
            if not isinstance(other, OPCLIVersion):
                other = OPCLIVersion(other)

            parts_self, parts_other = self._normalize(other)
            equal = True
            for i, part in enumerate(parts_self):
                if part != parts_other[i]:
                    equal = False
                    break

        return equal

    def __ne__(self, other):
        ne = not self.__eq__(other)
        return ne

    def __lt__(self, other):
        if id(self) == id(other):
            return False

        if not isinstance(other, type(self)):
            other = OPCLIVersion(other)

        lt = False
        _parts_self, _parts_other = self._normalize(other)

        for i, part in enumerate(_parts_self):
            if part > _parts_other[i]:
                break
            elif part == _parts_other[i]:
                continue
            elif part < _parts_other[i]:
                lt = True
                break

        return lt

    def __le__(self, other):
        le = self.__lt__(other) or self.__eq__(other)
        return le


MINIMUM_SERVICE_ACCOUNT_VERSION = OPCLIVersion('2.18.0-beta.01')
DOCUMENT_BYTES_BUG_VERSION = OPCLIVersion('2.2.0')



from .._py_op_commands import (
    EXISTING_AUTH_AVAIL,
    EXISTING_AUTH_IGNORE,
    EXISTING_AUTH_REQD,
    ExistingAuthEnum
)

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = [
    "EXISTING_AUTH_AVAIL",
    "EXISTING_AUTH_IGNORE",
    "EXISTING_AUTH_REQD",
    "ExistingAuthEnum"
]



# expect more recipe constants to be added over time
from ..op_items.password_recipe import (
    LETTERS_DIGITS_25,
    LETTERS_DIGITS_SYMBOLS_20
)

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = [
    "LETTERS_DIGITS_25",
    "LETTERS_DIGITS_SYMBOLS_20"
]



"""
Various exception classes raised by ponepassword API
TODO: Move other exception classes here
"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, List, Optional

from ._py_op_deprecation import deprecated

if TYPE_CHECKING:
    from .op_items._item_list import OPItemList  # pragma: no coverage


class OPBaseException(Exception):

    def __init__(self, msg):
        super().__init__(msg)


class _OPCalledProcessException(OPBaseException):
    """
    Base class to represent 'op' command failure.
    No code should handle this exception directly.

    Override this class and set MSG
    """
    MSG: Optional[str] = None

    def __init__(self, stderr_out, returncode):
        if not self.MSG:  # pragma: no coverage
            raise Exception(
                "subclass _OPCalledProcessException and override MSG")
        super().__init__(self.MSG)
        self.err_output = stderr_out
        self.returncode = returncode


class OPCmdFailedException(_OPCalledProcessException):
    """
    Generic Exception class for when an `op` command fails.

    Description:
    Raised from subprocess call-site when `op` executable returns non-zero

    Caller should handle this exception and raise a more descriptive exception reflecting
    the action that failed:

    Example:
    try:
        self._run(argv, capture_stdout=True, input_string=password)
    except OPCmdFailedException as ocfe:
        raise OPSigninException.from_opexception(ocfe) from ocfe
    """
    MSG = "'op' command failed"

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)

    @classmethod
    def from_opexception(cls, ope: OPCmdFailedException):
        return cls(ope.err_output, ope.returncode)


class OPCLIPanicException(_OPCalledProcessException):
    """
    Occasionally we're able to trigger a panic in the go runtime when executing 'op'

    When that happens, we want to raise a special exception in order to report it

    This class intentionally mirrors but does not extend OPCmdFailedException. We
    don't want to mask any 'op' crashes with code that handles OPCmdFailedException
    """
    MSG = "1Password CLI command crashed"

    def __init__(self, stderr_out: str, returncode: int, argv: List[str]):
        super().__init__(stderr_out, returncode)
        self.argv = list(argv)


class OPSigninException(OPCmdFailedException):
    MSG = "1Password sign-in failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPSignoutException(OPCmdFailedException):  # pragma: no coverage
    MSG = "1Password signout failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPForgetException(OPCmdFailedException):  # pragma: no coverage
    MSG = "1Password forget failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPItemGetException(OPCmdFailedException):
    MSG = "1Password 'get item' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPItemListException(OPCmdFailedException):
    MSG = "1Password 'item list' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPItemDeleteException(OPCmdFailedException):
    MSG = "1Password 'item delete' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPItemDeleteMultipleException(OPItemDeleteException):
    """
    Exception class for OP.item_delete_multiple() failures

    Contains list of deleted items, if any, before the failure occured
    """
    MSG = "1Password 'item delete' for multiple objects failed."

    def __init__(self, deleted_items: OPItemList, stderr_out, returncode):
        """
        Parameters
        ----------
        deleted_items : OPItemList
            List of items that were deleted before the failure occured
        stderr_out : str
            Error output from the 'op' command
        returncode : int
            Exit status of the 'op' op command
        """
        self.deleted_items = deleted_items
        super().__init__(stderr_out, returncode)

    @classmethod
    # type: ignore[override]
    def from_opexception(cls, ope: OPCmdFailedException, deleted_items: OPItemList):
        return cls(deleted_items, ope.err_output, ope.returncode)


class OPDocumentGetException(OPCmdFailedException):
    MSG = "1Password 'document get' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPDocumentDeleteException(OPCmdFailedException):
    MSG = "1Password 'document delete' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPUserGetException(OPCmdFailedException):
    MSG = "1Password 'get user' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPUserListException(OPCmdFailedException):
    MSG = "1Password 'user list' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPVaultGetException(OPCmdFailedException):
    MSG = "1Password 'get vault' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPVaultListException(OPCmdFailedException):
    MSG = "1Password 'vault list' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPGroupGetException(OPCmdFailedException):
    MSG = "1Password 'get group' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPGroupListException(OPCmdFailedException):
    MSG = "1Password 'group list' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPItemCreateException(OPCmdFailedException):  # pragma: no coverage
    MSG = "1Password 'item create' failed."

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPWhoAmiException(OPCmdFailedException):
    MSG = "1Password 'whoami' failed"

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPCmdMalformedSvcAcctTokenException(OPCmdFailedException):  # pragma: no coverage
    """
    The 'op' CLI command failed to parse a service account token
    """
    # Although raised from OP._whoami(), this shouldn't extend OPWhoAmiException
    # callers need to be able to catch OPWhoAmiException independently of this
    # exception
    MSG = "1Password 'whoami' failed due to malformed service account token"

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPRevokedSvcAcctTokenException(_OPCalledProcessException):
    MSG = "1Password operation failed due to a revoked service account token"

    def __init__(self, stderr_out, returncode):
        super().__init__(stderr_out, returncode)


class OPInvalidItemException(OPBaseException):
    def __init__(self, msg):
        super().__init__(msg)


@deprecated("Use OPAuthenticationException")
class OPNotSignedInException(OPBaseException):
    """
    DEPRECATION NOTE:
    This exception class is deprecated in favor of OPAuthenticationException
    and will be removed in a future version

    Exception indicating the `op` command is not authenticated or is unable to complete
    authentication
    """

    def __init__(self, msg):
        super().__init__(msg)


class OPAuthenticationException(OPNotSignedInException):
    # TODO: inherit from OPBaseException once
    # OPNotSignedInException removed
    """
    Exception indicating the `op` command is not authenticated or is unable to complete
    authentication
    """
    _skip_drecation_warn = True

    def __init__(self, msg):
        super().__init__(msg)


class OPInvalidDocumentException(OPInvalidItemException):
    def __init__(self, msg):
        super().__init__(msg)


class OPNotFoundException(Exception):
    MSG = "1Password cli command not found at path: %s"

    def __init__(self, op_path, errno):
        msg = self.MSG % op_path
        self.errno = errno
        super().__init__(msg)


class OPConfigNotFoundException(Exception):
    pass


class OPInvalidFieldException(OPBaseException):
    def __init__(self, msg):
        super().__init__(msg)


class OPUnknownAccountException(OPBaseException):
    def __init__(self, msg):
        super().__init__(msg)


_deprecated_exceptions = {
    OPNotSignedInException.__name__: OPAuthenticationException.__name__
}

# replace OPNotSignedInException with _OPNotSignedInException
# in order to trigger deprecation warnings
_OPNotSignedInException = OPNotSignedInException
del OPNotSignedInException


def __getattr__(name: str):
    # module level __getattr__() is valid as of python 3.7, pep-562
    # handling deprecation warnings on import is a key use-case
    # https://peps.python.org/pep-0562/
    if name in _deprecated_exceptions:
        _deprecated_name = f"_{name}"
        alternate = _deprecated_exceptions[name]
        warnings.warn(
            f"Exception class {name} is deprecated. Use {alternate}", category=FutureWarning)
        return globals()[_deprecated_name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor

# {
#     "uuid": "zjc6s5ri3rhcxploofa67jamze",
#     "templateUuid": "003",
#     "trashed": "N",
#     "createdAt": "2021-03-19T23:27:12Z",
#     "updatedAt": "2021-03-19T23:30:10Z",
#     "changerUuid": "5GHHPJK5HZC5BAT7WDUXW57G44",
#     "itemVersion": 2,
#     "vaultUuid": "gshlsjsajnawtnjynzgwmiebge",
#     "details": {
#         "notesPlain": "Note text here. **Mardown** supported.\n\nWhat does the note text look like?",
#         "passwordHistory": [],
#         "sections": [
#             {
#                 "name": "linked items",
#                 "title": "Related Items"
#             }
#         ]
#     },
#     "overview": {
#         "ainfo": "Note text here. **Mardown** supported.",
#         "ps": 0,
#         "title": "Example Secure Note"
#     }
# }


@op_register_item_descriptor_type
class OPSecureNoteItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "SECURE_NOTE"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPSecureNoteItem(OPAbstractItem):
    CATEGORY = "SECURE_NOTE"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def note_text(self):
        text = self.field_value_by_id("notesPlain")
        return text


@op_register_item_type
class OPSecureNoteItemRelaxedValidation(OPSecureNoteItem):
    _relaxed_validation = True



from ..op_items.item_types._item_descriptor_base import (
    OPAbstractItemDescriptor
)
from ..op_items.item_types.api_credential import OPAPICredentialItemDescriptor
from ..op_items.item_types.credit_card import OPCreditCardItemDescriptor
from ..op_items.item_types.document import OPDocumentItemDescriptor
from ..op_items.item_types.identity import OPIdentityItemDescriptor
from ..op_items.item_types.login import OPLoginDescriptorItem
from ..op_items.item_types.password import OPPasswordItemDescriptor
from ..op_items.item_types.secure_note import OPSecureNoteItemDescriptor
from ..op_items.item_types.server import OPServerItemDescriptor
from ..op_items.item_types.ssh_key import OPSSHKeyItemDescriptor
from ..op_objects import (
    OPGroupDescriptor,
    OPGroupDescriptorList,
    OPUserDescriptor,
    OPUserDescriptorList,
    OPVaultDescriptor,
    OPVaultDescriptorList
)

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = ["OPGroupDescriptor",
           "OPGroupDescriptorList",
           "OPUserDescriptor",
           "OPUserDescriptorList",
           "OPVaultDescriptor",
           "OPVaultDescriptorList",
           "OPAbstractItemDescriptor",
           "OPAPICredentialItemDescriptor",
           "OPCreditCardItemDescriptor",
           "OPDocumentItemDescriptor",
           "OPIdentityItemDescriptor",
           "OPLoginDescriptorItem",
           "OPPasswordItemDescriptor",
           "OPSecureNoteItemDescriptor",
           "OPServerItemDescriptor",
           "OPSSHKeyItemDescriptor"]



from typing import Dict, List, Optional, Union

from ..._abc_meta import enforcedmethod
from ...py_op_exceptions import OPInvalidItemException
from ..field_registry import OPItemFieldFactory
from ..fields_sections.item_field_base import OPItemField
from ..fields_sections.item_section import (
    OPItemFieldCollisionException,
    OPSection,
    OPSectionCollisionException
)
from ..item_validation_policy import get_relaxed_validation
from ._item_descriptor_base import OPAbstractItemDescriptor


class OPSectionNotFoundException(Exception):
    pass


class OPFieldNotFoundException(Exception):
    pass


class OPAbstractItem(OPAbstractItemDescriptor):
    CATEGORY: Optional[str] = None
    FROM_TEMPLATE = False
    # relaxed validation is disabled by default
    # may be overridden in a "relaxed" sublcass
    # E.g. OPLoginItemRelaxedValidation
    _relaxed_validation: bool = False

    @enforcedmethod
    def __init__(self, item_dict_or_json: Union[Dict, str]):
        super().__init__(item_dict_or_json)
        self._section_map = self._initialize_sections()
        self._field_map = self._initialize_fields()

    @property
    def sections(self) -> List[OPSection]:
        section_list = self.get("sections", [])
        return section_list

    def relaxed_validation(self) -> bool:
        """
        Get relaxed validation policy

        May be determined by any of the following:
        - This class has relaxed validation enabled (e.g., OPLoginItemRelaxedValidation)
        - Relaxed validation is globally set for this class via OPItemValidationPolicy
        - Relaxed validation is globally set for all op item classes

        Returns
        -------
        bool
            Whether relaxed validation policy is enabled
        """
        relaxed = False
        if not self.FROM_TEMPLATE:
            relaxed = self._relaxed_validation
            if not relaxed:
                relaxed = get_relaxed_validation(item_class=self.__class__)
        return relaxed

    def sections_by_label(self, label, case_sensitive=True) -> List[OPSection]:
        """
        Returns a list of zero or more sections matching the given title.
        Sections are not required to have unique titles, so there may be more than one match.
        """
        matching_sections = []
        sect: OPSection
        for sect in self.sections:
            s_label = sect.label
            if not case_sensitive:
                s_label = s_label.lower()
                label = label.lower()
            if s_label == label:
                matching_sections.append(sect)

        return matching_sections

    def section_by_id(self, section_id) -> OPSection:
        try:
            section: OPSection = self._section_map[section_id]
        except KeyError:
            raise OPSectionNotFoundException(
                f"Section not found with Section ID: {section_id}")

        return section

    def first_section_by_label(self, label, case_sensitive=True) -> Optional[OPSection]:
        sections = self.sections_by_label(label, case_sensitive=case_sensitive)
        section = None
        if sections:
            section = sections[0]
        return section

    def field_value_by_section_title(self, section_title: str, field_label: str):
        section = self.first_section_by_label(section_title)
        value = None
        if section is not None:
            value = self._field_value_from_section(section, field_label)
        return value

    def field_by_id(self, field_id) -> OPItemField:
        try:
            field = self._field_map[field_id]
        except KeyError:
            raise OPFieldNotFoundException(
                f"Field not found with ID: {field_id}")
        return field

    def fields_by_label(self, field_label: str, case_sensitive=True) -> List[OPItemField]:
        fields = []
        f: OPItemField
        for _, f in self._field_map.items():
            f_label = f.label
            if not case_sensitive:
                f_label = f_label.lower()
                field_label = field_label.lower()

            if f_label == field_label:
                fields.append(f)
        return fields

    def first_field_by_label(self, field_label: str, case_sensitive=True) -> OPItemField:
        fields = self.fields_by_label(
            field_label, case_sensitive=case_sensitive)
        f = fields[0]
        return f

    def field_value_by_id(self, field_id):
        field = self.field_by_id(field_id)
        value = field.value
        return value

    def field_reference_by_id(self, field_id) -> Optional[str]:
        field = self.field_by_id(field_id)
        ref = field.reference
        return ref

    def _field_value_from_section(self, section: OPSection, field_label: str):
        section_field: OPItemField = section.fields_by_label(field_label)[0]
        value = section_field.value
        return value

    def _initialize_sections(self):
        section_list = []
        section_map = {}
        _sections = self.get("sections")
        if _sections:
            for section_dict in _sections:
                s = OPSection(section_dict)
                try:
                    section_id = s.section_id
                except KeyError as e:
                    # in rare instances sections may lack an "id" altogether
                    # Let's treat that as having an empty string instead, since we have
                    # logic to deal with that below, and sometimes that occurs as well
                    if self.relaxed_validation():
                        section_id = ""
                    else:
                        raise OPInvalidItemException(
                            f"section has no ID {s}") from e

                if section_id in section_map:
                    # NOTE: in rare cases 'op' will return items
                    # that have multiple duplicated sections, including section ID
                    # resulting in section ID collisions. In these cases
                    # relaxed validation is required to parse the dictionary
                    # and return a usable object
                    # see:
                    # tests/test_non_conformant_data/test_duplicate_sections.py
                    # if relaxed validation is not enabled
                    # raise an exception
                    if not self.relaxed_validation():
                        raise OPSectionCollisionException(
                            f"Section {section_id} already registered")
                    else:
                        # for code coverage visibility
                        pass
                section_map[section_id] = s
                section_list.append(s)
        self["sections"] = section_list
        return section_map

    def _initialize_fields(self):
        relaxed_validation = self.relaxed_validation()
        field_list = []
        field_map = {}
        _fields = self.get("fields", [])
        for field_dict in _fields:
            field = OPItemFieldFactory.item_field(field_dict)
            try:
                field_id = field.field_id
            except KeyError as e:
                # Based on evidence of sections occasionally lacking "id" elements,
                # in theory there may be instances where fields also lack an "id"
                # Let's treat that as having an empty string instead, since we have
                # logic to deal with that below, and sometimes that occurs as well
                if self.relaxed_validation():
                    field_id = ""
                else:
                    raise OPInvalidItemException(
                        f"Field has no ID: {field.label}") from e
            if field_id in field_map:
                # NOTE: in many cases 'op' will return items
                # that have multiple fields with empty-string IDs
                # resulting in field ID collisions. In these cases
                # relaxed validation is required to parse the dictionary
                # and return a usable object

                # if relaxed validation is not enabled
                # raise an exception
                if not relaxed_validation:
                    raise OPItemFieldCollisionException(
                        f"Field {field_id} already registered")
                else:
                    # for code coverage visibility
                    pass
            section_dict = field.get("section")
            if section_dict:
                section_id = section_dict["id"]
                section = self.section_by_id(section_id)
                section.register_field(
                    field_dict, relaxed_validation=relaxed_validation)
            field_list.append(field)
            field_map[field_id] = field
        self["fields"] = field_list
        return field_map



from typing import List

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor


class OPDocumentFile(dict):
    def __init__(self, file_dict):
        super().__init__(file_dict)

    @property
    def file_id(self) -> str:
        return self["id"]

    @property
    def name(self) -> str:
        return self["name"]

    @property
    def size(self) -> int:
        return self["size"]

    @property
    def content_path(self) -> str:
        return self["content_path"]


@op_register_item_descriptor_type
class OPDocumentItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "DOCUMENT"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPDocumentItem(OPAbstractItem):
    CATEGORY = "DOCUMENT"

    def __init__(self, item_dict):
        super().__init__(item_dict)
        files = []
        for file_dict in self.files:
            file_obj = OPDocumentFile(file_dict)
            files.append(file_obj)
        self["files"] = files

    @property
    def file_name(self):
        first_file = self.files[0]
        file_name = first_file.name
        return file_name

    @property
    def files(self) -> List[OPDocumentFile]:
        return self["files"]


@op_register_item_type
class OPDocumentItemRelaxedValidation(OPDocumentItem):
    _relaxed_validation = True



from typing import Dict, Union

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ..fields_sections.item_field_base import OPItemField
from ..fields_sections.item_section import OPSection
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPCreditCardItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "CREDIT_CARD"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPCreditCardItem(OPAbstractItem):
    CATEGORY = "CREDIT_CARD"

    def __init__(self, item_dict: Union[Dict, str]):
        super().__init__(item_dict)

    def additional_details(self) -> OPSection:
        details = self.sections_by_label("Additional Details")[0]
        return details

    def addl_details_item(self, field_label: str):
        details = self.additional_details()
        item_value = self._details_item(details, field_label)
        return item_value

    @property
    def credit_card_number(self):
        ccnum = self.field_value_by_id("ccnum")
        return ccnum

    @property
    def cvv(self):
        ccv = self.field_value_by_id("cvv")
        return ccv

    @property
    def expiry_date(self) -> int:
        exp_date = self.field_value_by_id("expiry")
        exp_date = int(exp_date)
        return exp_date

    @property
    def valid_from(self) -> int:
        valid_from = self.field_value_by_id("validFrom")
        valid_from = int(valid_from)
        return valid_from

    @property
    def pin(self):
        pin = self.field_value_by_id("pin")
        return pin

    @property
    def credit_limit(self):
        climit = self.addl_details_item("credit limit")
        return climit

    @property
    def cash_withdrawal_limit(self):
        cw_limit = self.addl_details_item("cash withdrawal limit")
        return cw_limit

    @property
    def interest_rate(self):
        int_rate = self.addl_details_item("interest rate")
        return int_rate

    @property
    def issue_number(self):
        issue_num = self.addl_details_item("issue number")
        return issue_num

    def _details_item(self, details: OPSection, field_label: str):
        item_field: OPItemField = details.fields_by_label(field_label)[0]
        item_value = item_field.value
        return item_value


@op_register_item_type
class OPCreditCardItemRelaxedValidation(OPCreditCardItem):
    _relaxed_validation = True



"""
Miscellaneous classes for objects return by 'op get' other than item or document objects
"""
import json
from datetime import datetime
from json.decoder import JSONDecodeError
from typing import Dict, List, TypeVar, Union

from ._abc_meta import ABCMetaDict, enforcedmethod
from ._datetime import fromisoformat_z
from .json import safe_unjson
from .py_op_exceptions import OPBaseException


class OPInvalidObjectException(OPBaseException):
    """
    The data provided to generate an OP query object failed to parse or validate

    Attributes
    ----------
    object_json : Union[str, None]
        The original JSON from the 'op' command query, if available
    """

    def __init__(self, msg, object_json):
        super().__init__(msg)
        self.object_json = object_json


class OPInvalidUserException(OPInvalidObjectException):
    """
    The data provided to generate an 'op get user' query object failed to parse or validate
    """
    pass


class OPInvalidGroupException(OPInvalidObjectException):
    """
    The data provided to from 'op get group' failed to parse or validate
    """
    pass


class OPInvalidUserListException(OPInvalidObjectException):
    """
    The data provided from 'op list users' failed to parse or validate
    """
    pass


class OPInvalidGroupListException(OPInvalidObjectException):
    """
    The data provided from 'op list groups' failed to parse or validate
    """
    pass


class OPInvalidVaultException(OPInvalidObjectException):
    """
    The data provided from 'op get vault' failed to parse or validate
    """
    pass


class OPInvalidVaultListException(OPInvalidGroupListException):
    """
    The data provided from 'op list vaults' failed to parse or validate
    """


DT = TypeVar('DT')


class _OPDescriptorList(List[DT]):
    """
    A class for type-hinting lists of 1Password object descriptors
    """
    pass


class OPAbstractObject(dict, metaclass=ABCMetaDict):
    """
    Abstract base class for 1Password objects as returned by 'op list|get <object class>'
    """

    @enforcedmethod
    def __init__(self, dict_or_json: Union[str, dict]):
        obj_dict = safe_unjson(dict_or_json)
        super().__init__(obj_dict)

    @property
    def unique_id(self):
        """
        str: The object's "id" field
        """
        return self["id"]

    @property
    def name(self) -> str:
        """
        str : The name attribute
        """
        return self["name"]


class OPBaseObject(OPAbstractObject):
    """
    Base class for miscellaneous 1Password objects as returned by 'op get <object class>'
    """

    @property
    def created_at(self) -> datetime:
        """
        datetime : The created_at attribute parsed as a datetime object
        """
        created = self["created_at"]
        created = fromisoformat_z(created)
        return created

    @property
    def updated_at(self) -> datetime:
        """
        datetime : The updated_at attribute parsed as a datetime object
        """
        updated = self["updated_at"]
        updated = fromisoformat_z(updated)
        return updated


class OPUser(OPBaseObject):
    """
    A dictionary of full details about a 1Password user.
    This can be treated as a normal dictionary. In addition, it has a convenience
    property for each key in the the dictionary.

    Note
    ----
    Date-related properties return parsed 'datetime' objects. To access the original
    date strings, use the corresponding dictionary key.
    """

    def __init__(self, user_dict_or_json: Union[str, dict]):
        """
        Parameters
        ----------
        user_dict_or_json : Union[str, dict]
            A dictionary or JSON string return from 'op get user'. If a JSON string is provided,
            it will first be unserialized to a dict object.

        Raises
        ------
        OPInvalidUserException
            If JSON is provided and unserializing fails.

        """
        try:
            super().__init__(user_dict_or_json)
        except JSONDecodeError as jdce:
            raise OPInvalidUserException(
                f"Failed to unserialize user json: {jdce}", user_dict_or_json) from jdce

    @property
    def email(self) -> str:
        """
        str : The email attribute
        """
        return self["email"]

    @property
    def type(self) -> str:
        """
        str : The type attribute
        """
        return self["type"]

    @property
    def state(self) -> str:
        """
        str : The state attribute
        """
        return self["state"]

    @property
    def last_auth_at(self) -> datetime:
        """
        datetime : The lastAuthAt attribute parsed as a datetime object
        """
        last_auth = self["last_auth_at"]
        last_auth = fromisoformat_z(last_auth)
        return last_auth


class OPGroup(OPBaseObject):
    """
    A dictionary of full details about a 1Password group.
    This can be treated as a normal dictionary. In addition, it has a convenience
    property for each key in the the dictionary.

    Note
    ----
    Date-related properties return parsed 'datetime' objects. To access the original
    date strings, use the corresponding dictionary key.
    """

    def __init__(self, group_dict_or_json: Union[str, dict]):
        """
        Parameters
        ----------
        user_dict_or_json : Union[str, dict]
            A dictionary or JSON string return from 'op get user'. If a JSON string is provided,
            it will first be unserialized to a dict object.

        Raises
        ------
        OPInvalidGroupException
            If JSON is provided and unserializing fails.

        """
        try:
            super().__init__(group_dict_or_json)
        except JSONDecodeError as jdce:
            raise OPInvalidGroupException(
                f"Failed to unserialize group json: {jdce}", group_dict_or_json) from jdce

    @property
    def description(self) -> str:
        """
        The desc attribute
        """
        return self["description"]

    @property
    def state(self) -> str:
        return self["state"]

    @property
    def permissions(self) -> List[str]:
        return self["permissions"]

    @property
    def type(self) -> str:
        """
        str : The type attribute
        """
        return self["type"]


class OPUserDescriptor(OPAbstractObject):
    """
    A dictionary describing a user as returned by 'op list users'. This is a subset of a full OPUserOboject.
    """

    def __init__(self, user_descriptor_dict: Dict):
        super().__init__(user_descriptor_dict)

    @property
    def email(self) -> str:
        """
        str : The email attribute
        """
        return self["email"]

    @property
    def type(self) -> str:
        """
        str : The type attribute
        """
        return self["type"]

    @property
    def state(self) -> str:
        """
        str : The state attribute
        """
        return self["state"]


class OPUserDescriptorList(_OPDescriptorList[OPUserDescriptor]):
    """
    List of OPUserDescriptor dictionaries as returned from an 'op list users' operation
    These are not full user dictionaries as would be returned from 'op get user'.
    Rather, just a minimum set of fields to describe a user.

    Each user descriptor object has the following properties
    """

    def __init__(self, user_list_json):
        super().__init__()
        user_list = []
        try:
            user_list = json.loads(user_list_json)
        except JSONDecodeError as jdce:
            raise OPInvalidUserListException(
                f"Failed to unserialize user json: {jdce}", user_list_json)
        for user_obj in user_list:
            user_descriptor = OPUserDescriptor(user_obj)
            self.append(user_descriptor)


class OPGroupDescriptor(OPAbstractObject):
    """
    A dictionary describing a group as returned by 'op list groups'. This is a subset of a full OPGroupOboject.
    """

    # Override __init__ since parent class is abstract
    def __init__(self, dict_or_json: Union[str, dict]):
        super().__init__(dict_or_json)

    @property
    def description(self) -> str:
        """
        The desc attribute
        """
        return self["description"]

    @property
    def state(self) -> str:
        return self["state"]

    @property
    def created_at(self) -> datetime:
        """
        datetime : The created_at attribute parsed as a datetime object
        """
        created = self["created_at"]
        created = fromisoformat_z(created)
        return created


class OPGroupDescriptorList(_OPDescriptorList[OPGroupDescriptor]):
    """
    List of OPGroupDescriptor dictionaries as returned from an 'op list group' operation.
    These are not full group dictionaries as would be returned from 'op get group'.
    Rather, just a minimum set of fields to describe a group.

    Each group descriptor is a dictionary of str:str pairs, but in addition has the following
    convenience properties
    """

    def __init__(self, group_list_json):
        super().__init__()
        group_list = []
        try:
            group_list = json.loads(group_list_json)
        except JSONDecodeError as jdce:
            raise OPInvalidGroupListException(
                f"Failed to unserialize user json: {jdce}", group_list_json) from jdce
        for group_obj in group_list:
            group_descriptor = OPGroupDescriptor(group_obj)
            self.append(group_descriptor)


class OPVaultDescriptor(OPAbstractObject):
    """
    A dictionary describing a vault as returned by 'op list vaults'. This is a subset of
    a full vault object.
    """

    def __init__(self, vault_dict_or_json):
        super().__init__(vault_dict_or_json)


class OPVault(OPVaultDescriptor):
    """
    A dictionary of full details about a vault as returned by
    an 'op get vault' operation
    """

    def __init__(self, vault_dict_or_json):
        try:
            super().__init__(vault_dict_or_json)
        except JSONDecodeError as jdce:
            raise OPInvalidVaultException(
                f"Failed to unserialize vault json: {jdce}", vault_dict_or_json) from jdce

    @property
    def description(self) -> str:
        """
        str : The vault description
        """
        return self["description"]

    @property
    def item_count(self) -> int:
        return self["items"]

    @property
    def type(self) -> str:
        """
        str: The "type" attribute of the vault.
        """
        return self["type"]

    @property
    def created_at(self) -> datetime:
        """
        datetime : The created_at attribute parsed as a datetime object
        """
        created = self["created_at"]
        created = fromisoformat_z(created)
        return created

    @property
    def updated_at(self) -> datetime:
        """
        datetime : The updated_at attribute parsed as a datetime object
        """
        updated = self["updated_at"]
        updated = fromisoformat_z(updated)
        return updated


class OPVaultDescriptorList(_OPDescriptorList[OPVaultDescriptor]):
    """
    A list of OPVaultDescriptor dictionaries as returned by 'op list vaults'
    """

    def __init__(self, vault_list_json: str):
        super().__init__()
        vault_list = []
        try:
            vault_list = json.loads(vault_list_json)
        except JSONDecodeError as jdce:
            raise OPInvalidVaultListException(
                f"Failed to unserialize vault list JSON: {jdce}", vault_list_json) from jdce

        for vault_obj in vault_list:
            vault_descriptor = OPVaultDescriptor(vault_obj)
            self.append(vault_descriptor)



import logging

DEBUG = logging.DEBUG
INFO = logging.INFO


def console_logger(name: str, level: int):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    logger.addHandler(ch)

    return logger



from typing import Optional

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem, OPFieldNotFoundException
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPServerItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "SERVER"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPServerItem(OPAbstractItem):
    CATEGORY = "SERVER"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def password(self) -> Optional[str]:
        password = self.field_value_by_id("password")
        return password

    @property
    def username(self) -> Optional[str]:
        username = self.field_value_by_id("username")
        return username

    @property
    def url(self) -> Optional[str]:
        try:
            url = self.field_value_by_id("url")
        except OPFieldNotFoundException:
            url = None
        return url

    @property
    def admin_console_password(self) -> Optional[str]:
        password = self.field_value_by_section_title(
            "Admin Console", "console password")
        return password

    @property
    def admin_console_username(self) -> Optional[str]:
        username = self.field_value_by_section_title(
            "Admin Console", "admin console username")
        return username

    @property
    def admin_console_url(self) -> Optional[str]:
        url = self.field_value_by_section_title(
            "Admin Console", "admin console URL")
        return url

    @property
    def hosting_provider_name(self) -> Optional[str]:
        """
        Convenience property to retrieve the hosting provider name

        Returns
        -------
        str
            name of the hosting provider
        """
        # Note: the actual field ID and label are "name"
        # so we look it up by section to be sure
        return self.field_value_by_section_title("Hosting Provider", "name")

    @property
    def hosting_provider_website(self) -> Optional[str]:
        """
        Convenience property to retrieve the hosting provider website

        Returns
        -------
        str
            Value of the hosting provider website field
        """
        # Note: the actual field ID and label are "name"
        # so we look it up by section to be sure
        return self.field_value_by_section_title("Hosting Provider", "website")

    @property
    def support_contact_url(self) -> Optional[str]:
        """
        Convenience property to retrieve the hosting provider support URL

        Returns
        -------
        str
            Value of the hosting provider support URL field
        """
        url = self.field_value_by_id("support_contact_url")
        return url

    @property
    def support_contact_phone(self) -> Optional[str]:
        """
        Convenience property to retrieve the hosting provider support phone

        Returns
        -------
        str
            Value of the hosting provider support phone field
        """
        phone = self.field_value_by_id("support_contact_phone")
        return phone


@op_register_item_type
class OPServerItemRelaxedValidation(OPServerItem):
    _relaxed_validation = True



from typing import Any, Union


class OPItemField(dict):
    FIELD_TYPE: Union[str, None] = None

    def __init__(self, field_dict):
        super().__init__(field_dict)

    @property
    def field_id(self) -> str:
        return self["id"]

    @property
    def label(self) -> str:
        """
        Returns the field label as assigned and seen in the 1Password UI
        """
        return self["label"]

    @property
    def value(self) -> Any:
        """
        Returns the field's value (password, URL, etc.) as assigned and seen in the 1Password UI,
        or None if the field lacks a value
        """
        v = self.get("value")
        return v

    @property
    def field_type(self) -> str:
        """
        Returns the field's type, which affects how the field's value is rendered in 1Password
        """
        return self["type"]

    @property
    def reference(self) -> Union[str, None]:
        return self.get("reference")

    @property
    def purpose(self) -> str:
        return self["purpose"]

    @property
    def entropy(self) -> Union[float, None]:
        return self.get("entropy")

    @property
    def section_id(self) -> Union[str, None]:
        section_id = None
        section = self.get("section")
        if section:
            section_id = section["id"]
        return section_id



from ..field_registry import op_register_item_field_type
from .item_field_base import OPItemField


@op_register_item_field_type
class OPStringField(OPItemField):
    FIELD_TYPE = "STRING"


@op_register_item_field_type
class OPConcealedField(OPItemField):
    FIELD_TYPE = "CONCEALED"


@op_register_item_field_type
class OPTOTPField(OPStringField):
    FIELD_TYPE = "OTP"

    @property
    def totp_secret(self) -> str:
        return self.value

    @property
    def totp(self) -> str:
        return self["totp"]



import datetime
from typing import Dict, Union

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPAPICredentialItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "API_CREDENTIAL"

    def __init__(self, item_dict):
        super().__init__(item_dict)


"""
{
  "id": "jgou2ypa4ceurm4fetya3kptau",
  "title": "Example API Credential",
  "version": 3,
  "vault": {
    "id": "gshlsjsajnawtnjynzgwmiebge",
    "name": "Test Data"
  },
  "category": "API_CREDENTIAL",
  "last_edited_by": "5GHHPJK5HZC5BAT7WDUXW57G44",
  "created_at": "2022-05-04T01:08:14Z",
  "updated_at": "2022-05-04T01:09:50Z",
  "additional_information": "other",
  "fields": [
    {
      "id": "notesPlain",
      "type": "STRING",
      "purpose": "NOTES",
      "label": "notesPlain",
      "reference": "op://Test Data/Example API Credential/notesPlain"
    },
    {
      "id": "username",
      "type": "STRING",
      "label": "username",
      "value": "__token__",
      "reference": "op://Test Data/Example API Credential/username"
    },
    {
      "id": "credential",
      "type": "CONCEALED",
      "label": "credential",
      "value": "212784acd95ae1d67e8d513602e7b5acf60dbb0c811d9628a23151aafade72b5",
      "reference": "op://Test Data/Example API Credential/credential"
    },
    {
      "id": "type",
      "type": "MENU",
      "label": "type",
      "value": "other",
      "reference": "op://Test Data/Example API Credential/type"
    },
    {
      "id": "filename",
      "type": "STRING",
      "label": "filename",
      "reference": "op://Test Data/Example API Credential/filename"
    },
    {
      "id": "validFrom",
      "type": "DATE",
      "label": "valid from",
      "value": "1641038460",
      "reference": "op://Test Data/Example API Credential/valid from"
    },
    {
      "id": "expires",
      "type": "DATE",
      "label": "expires",
      "value": "1656590460",
      "reference": "op://Test Data/Example API Credential/expires"
    },
    {
      "id": "hostname",
      "type": "STRING",
      "label": "hostname",
      "value": "api.example.com",
      "reference": "op://Test Data/Example API Credential/hostname"
    }
  ]
}
"""


@op_register_item_type
class OPAPICredentialItem(OPAbstractItem):
    CATEGORY = "API_CREDENTIAL"

    def __init__(self, item_dict: Union[Dict, str]):
        super().__init__(item_dict)

    @property
    def username(self):
        username = self.field_value_by_id("username")
        return username

    @property
    def credential(self) -> str:
        cred = self.field_value_by_id("credential")
        return cred

    @property
    def type(self) -> str:
        return self.field_value_by_id("type")

    @property
    def filename(self) -> str:
        return self.field_value_by_id("filename")

    @property
    def valid_from(self) -> datetime.date:
        valid_from = self.field_value_by_id("validFrom")
        valid_from = int(valid_from)
        valid_from = datetime.date.fromtimestamp(valid_from)
        return valid_from

    @property
    def expires(self) -> datetime.date:
        expires = self.field_value_by_id("expires")
        expires = int(expires)
        expires = datetime.date.fromtimestamp(expires)
        return expires

    @property
    def hostname(self) -> str:
        return self.field_value_by_id("hostname")


@op_register_item_type
class OPAPICredentialItemRelaxedValidation(OPAPICredentialItem):
    _relaxed_validation = True



from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor

"""
This module contains classes that may optionally be used as stand-ins for unknown
item and item descriptor types

These classes are not exposed as API as they are not intended to be used directly

Rather, the caller should pass the appropriate flags to API methods to enable their use
"""


class _OPGenericItemDescriptor(OPAbstractItemDescriptor):
    """
    A generic item descriptor class that can optionally be used as a fallback
    when an item descriptor list contains an unknown item type
    """

    def __init__(self, item_dict_or_json):
        super().__init__(item_dict_or_json)


class _OPGenericItem(OPAbstractItem):  # pragma: no coverage
    """
    A generic item class that can optionally be used as a fallback when
    retrieving an item of unknown type
    """

    def __init__(self, item_dict_or_json):
        super().__init__(item_dict_or_json)


class _OPGenericItemRelaxedValidation(_OPGenericItem):
    """
    A relaxed validation of _OPGenericItem that will allow creation of item objects
    from non-conformant item dictionaries

    This class is not intended to be used directly (publically or internally). It
    is registered with
    """
    _relaxed_validation = True



# __future__.annotations, and typing.TYPE_CHECKING
# enable anything imported for type hinting to disappear at run time
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from pyonepassword.api.exceptions import OPItemDeleteException

# make imports for type-hinting disappear at run-time to avoid
# circular imports.
# this also reduced exercising tested code simply by importing
if TYPE_CHECKING:
    from pyonepassword import OP

    from .fixtures.expected_login import ExpectedLoginItemData


# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_item_delete_01(signed_in_op: OP, expected_login_item_data: ExpectedLoginItemData):
    """
    Test deleting an item based on its non-unique title
    """
    login_name = "Delete Me Unique"
    vault = "Test Data"
    expected = expected_login_item_data.data_for_login(login_name)
    expected_item_id = expected.unique_id
    result = signed_in_op.item_delete(login_name, vault=vault)
    assert result == expected_item_id


def test_item_delete_02(signed_in_op: OP, expected_login_item_data: ExpectedLoginItemData):
    """
    Test deleting and archiving an item based on its non-unique title
    """
    login_name = "Delete and Archive Me"
    vault = "Test Data"
    expected = expected_login_item_data.data_for_login(login_name)
    expected_item_id = expected.unique_id
    result = signed_in_op.item_delete(login_name, vault=vault, archive=True)
    assert result == expected_item_id


def test_item_delete_03(signed_in_op: OP, expected_login_item_data: ExpectedLoginItemData):
    """
    Test:
      - deleting and archiving an item based on its non-unique title
      - fetching the item with include_archive=True
    """
    login_name = "Delete and Archive Me"
    vault = "Test Data"
    expected = expected_login_item_data.data_for_login(login_name)
    expected_item_id = expected.unique_id
    signed_in_op.item_delete(login_name, vault=vault, archive=True)
    result = signed_in_op.item_get(login_name, include_archive=True)
    assert result.unique_id == expected_item_id


def test_item_delete_non_existent_01(signed_in_op: OP):
    """
    Test deleting a non-existent item
    """
    login_name = "non-existent-item"
    vault = "Test Data"

    with pytest.raises(OPItemDeleteException):
        signed_in_op.item_delete(login_name, vault=vault)


def test_item_delete_non_existent_02(signed_in_op: OP):
    """
    Test deleting a non-existent item, bypassing item_get()
    """
    login_name = "non-existent-item"
    vault = "Test Data"

    with pytest.raises(OPItemDeleteException):
        # OP.item_delete() calls item_get() first
        # which will fail on non-existent items
        # so in order to test delete operatation failing, we
        # need to call private ._item_delete() interface
        signed_in_op._item_delete(login_name, vault=vault)


def test_item_delete_duplicate_01(signed_in_op: OP):
    """
    Test:
      - deleting an item based on its non-unique title
      - when there are two or more items sharing the same title
    """
    login_name = "Delete Me"
    vault = "Test Data"

    with pytest.raises(OPItemDeleteException):
        signed_in_op.item_delete(login_name, vault=vault)



from ._item_type_registry import OPItemFactory
from .item_types._item_descriptor_base import OPAbstractItemDescriptor
from .item_types.generic_item import _OPGenericItemDescriptor


class OPItemDescriptorFactory(OPItemFactory):
    _GENERIC_ITEM_CLASS = _OPGenericItemDescriptor
    _TYPE_REGISTRY = {}

    @classmethod
    def item_descriptor(cls, item_json, generic_okay=False) -> OPAbstractItemDescriptor:
        obj = cls.op_item(item_json, generic_okay=generic_okay)
        return obj


def op_register_item_descriptor_type(item_class):
    item_type = item_class.CATEGORY
    OPItemDescriptorFactory.register_op_item_type(item_type, item_class)
    return item_class



import pytest

from pyonepassword import OP
from pyonepassword.api.exceptions import (
    OPCLIPanicException,
    OPNotFoundException,
    OPSigninException
)


@pytest.mark.usefixtures("valid_op_cli_config_homedir")
def test_missing_op():
    with pytest.raises(OPNotFoundException):
        OP(op_path="no-such-op")


@pytest.mark.usefixtures("valid_op_cli_config_homedir")
@pytest.mark.usefixtures("setup_normal_op_env_signin_failure")
def test_signin_fail():
    with pytest.raises(OPSigninException):
        OP(op_path="mock-op")


@pytest.mark.usefixtures("valid_op_cli_config_homedir")
@pytest.mark.usefixtures("setup_svc_acct_corrupt_env")
def test_op_cli_crash(console_logger):
    expected_argv = ['mock-op', '--format', 'json', 'whoami']
    try:
        OP(op_path="mock-op", logger=console_logger)
        assert False, "Should have raised OPCLIPanicException"
    except OPCLIPanicException as e:
        assert expected_argv == e.argv



import os

import pytest

from pyonepassword._op_cli_config import OPCLIConfig
from pyonepassword.api.exceptions import OPConfigNotFoundException

from .fixtures.expected_op_cli_config import ExpectedConfigData
from .fixtures.platform_support import DEV_NULL, HOME_ENV_VAR, is_windows


def _sanity_check_xdg_home_env():
    assert os.environ.get('XDG_CONFIG_HOME') is not None
    assert os.environ.get(HOME_ENV_VAR) in [DEV_NULL, None]


def _sanity_check_standard_home_env():
    assert os.environ.get('XDG_CONFIG_HOME') is None
    assert os.environ.get(HOME_ENV_VAR) not in [DEV_NULL, None]


def test_op_cli_config_homedir_01(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.shorthand == expected.shorthand


def test_op_cli_config_homedir_02(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.account_uuid == expected.account_uuid


def test_op_cli_config_homedir_03(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.url == expected.url


def test_op_cli_config_homedir_04(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.user_uuid == expected.user_uuid


def test_op_cli_config_homedir_05(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.email == expected.email


def test_op_cli_config_homedir_06(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.uuid_for_account("example_shorthand")
    assert result == expected.user_uuid


def test_op_cli_config_homedir_07(valid_op_cli_config_homedir):
    _sanity_check_standard_home_env()
    shorthand = "NO_SUCH_SHORTHAND"
    config = OPCLIConfig()
    with pytest.raises(OPConfigNotFoundException):
        config.get_config(shorthand)


@pytest.mark.usefixtures("valid_op_cli_config_no_shorthand")
def test_op_cli_config_alt_acct_identifiers_01(expected_op_config_data: ExpectedConfigData):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    user_uuid = "5GHHPJK5HZC5BAT7WDUXW57G44"
    config = OPCLIConfig()
    result = config.get_config(user_uuid)
    assert expected.user_uuid == result.user_uuid


@pytest.mark.usefixtures("valid_op_cli_config_no_shorthand")
def test_op_cli_config_alt_acct_identifiers_02(expected_op_config_data: ExpectedConfigData):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    user_email = "example_user@example.email"
    config = OPCLIConfig()
    result = config.get_config(user_email)
    assert expected.user_uuid == result.user_uuid


@pytest.mark.usefixtures("valid_op_cli_config_no_shorthand")
def test_op_cli_config_alt_acct_identifiers_03(expected_op_config_data: ExpectedConfigData):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    account_uuid = "GRXJAN4BY5DPROISKYL55IRCPY"
    config = OPCLIConfig()
    result = config.get_config(account_uuid)
    assert expected.user_uuid == result.user_uuid


@pytest.mark.usefixtures("valid_op_cli_config_no_shorthand")
def test_op_cli_config_alt_acct_identifiers_04(expected_op_config_data: ExpectedConfigData):
    _sanity_check_standard_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    account_url = "https://example-account.1password.com"
    config = OPCLIConfig()
    result = config.get_config(account_url)
    assert expected.user_uuid == result.user_uuid


def test_op_cli_config_xdg_01(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    _sanity_check_xdg_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.shorthand == expected.shorthand


def test_op_cli_config_xdg_02(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    _sanity_check_xdg_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.account_uuid == expected.account_uuid


def test_op_cli_config_xdg_03(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    _sanity_check_xdg_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.url == expected.url


def test_op_cli_config_xdg_04(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    _sanity_check_xdg_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.user_uuid == expected.user_uuid


def test_op_cli_config_xdg_05(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    _sanity_check_xdg_home_env()
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.get_config("example_shorthand")
    assert result.email == expected.email


def test_op_cli_config_xdg_06(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    expected = expected_op_config_data.data_for_key("example-account")
    config = OPCLIConfig()
    result = config.uuid_for_account("example_shorthand")
    assert result == expected.user_uuid


def test_op_cli_config_xdg_07(expected_op_config_data: ExpectedConfigData, valid_op_cli_config_xdghome):
    shorthand = "NO_SUCH_SHORTHAND"
    config = OPCLIConfig()
    with pytest.raises(OPConfigNotFoundException):
        config.get_config(shorthand)


def test_op_cli_config_unreable_01(invalid_op_cli_config_unreable):
    # NOTE: This test will fail if run as root (e.g., in a docker container with no users)
    # there is no way to make a file unreadable to root
    if not is_windows():
        # this test depends on creating a config file that's not readable
        # this is not straightforward on windows via native python APIs
        # so only run this test if not on windows
        with pytest.raises(OPConfigNotFoundException):
            OPCLIConfig()
    else:
        assert True


def test_op_cli_config_missing_01(invalid_op_cli_config_missing):
    with pytest.raises(OPConfigNotFoundException):
        OPCLIConfig()


def test_op_cli_config_missing_02(invalid_op_cli_config_missing):
    with pytest.raises(OPConfigNotFoundException):
        OPCLIConfig(configpath="no_such_path")


def test_op_cli_config_malformed_01(invalid_op_cli_config_malformed):
    with pytest.raises(OPConfigNotFoundException):
        OPCLIConfig()


def test_op_cli_config_missing_shorthand_01(valid_op_cli_config_no_shorthand):
    conf = OPCLIConfig()
    with pytest.raises(OPConfigNotFoundException):
        conf.get_config()


@pytest.mark.usefixtures("valid_op_cli_config_no_account_list")
# @pytest.mark.usefixtures("setup_normal_op_env")
def test_op_cli_config_no_account_list_01():
    """
    Verify we can instantiate OPCLIConfig() even when account list is null
    """
    OPCLIConfig()



import warnings


def test_api_deprecation_01():
    """
    Import deprecated OPNotSignedInException class from
    pyonepassword.api.exceptions

    Ensure there is a deprecation warning
    """
    with warnings.catch_warnings(record=True) as warning_list:
        from pyonepassword.api.exceptions import \
            OPNotSignedInException  # noqa: F401
        assert warning_list


def test_api_deprecation_02():
    """
    Import deprecated OPNotSignedInException class from
    pyonepassword.py_op_exceptions

    Ensure there is a deprecation warning
    """
    with warnings.catch_warnings(record=True) as warning_list:
        from pyonepassword.py_op_exceptions import \
            OPNotSignedInException  # noqa: F401
        assert warning_list



from json.decoder import JSONDecodeError
from typing import Any, Dict, Type, Union

from ..json import safe_unjson
from ..py_op_exceptions import OPInvalidItemException
from .item_types._item_base import OPAbstractItem
from .item_types._item_descriptor_base import OPAbstractItemDescriptor
from .item_types.generic_item import (
    _OPGenericItem,
    _OPGenericItemDescriptor,
    _OPGenericItemRelaxedValidation
)


class OPUnknownItemTypeException(Exception):
    def __init__(self, msg, item_dict=None):
        super().__init__(msg)
        self.item_dict = item_dict


GenericType = Union[Type[_OPGenericItemDescriptor],
                    Type[_OPGenericItem]]


class OPItemFactory:
    # Fallback class for when item is an unknown types and caller
    # has enabled generic items
    _GENERIC_ITEM_CLASS: GenericType = _OPGenericItem

    # Fallback class when caller has enabled generic items as well as
    # relaxed item validation
    _GENERIC_ITEM_CLASS_RELAXED_VALIDATION = _OPGenericItemRelaxedValidation
    # registry of item classes with strict validation
    _TYPE_REGISTRY: Dict[str, Type[OPAbstractItem]] = {}
    # registry of item classes with relaxed validation
    _RELAXED_TYPE_REGISTRY: Dict[str, Type[OPAbstractItem]] = {}

    @classmethod
    def register_op_item_type(cls, item_type: str, item_class: Type[OPAbstractItem]):
        """
        Register an OP item type class

        If the class has the _relaxed_validation attribute set to True, it will be added
        to the registry of relaxed validation classes, otherwise it is added to the
        normal registry

        Parameters
        ----------
        item_type : str
            The CATEGORY attribute of the class representing the item type, such as "LOGIN"
        item_class : Type[OPAbstractItem]
            The item type class to register

        Raises
        ------
        Exception
            If two item type classes with the same CATEGORY are registered
        """
        if getattr(item_class, "_relaxed_validation", False):
            registry = cls._RELAXED_TYPE_REGISTRY
        else:
            registry = cls._TYPE_REGISTRY

        if item_type in registry:
            raise Exception(  # pragma: no coverage
                f"duplicate for item type {item_type}: {item_class}")
        registry[item_type] = item_class

    @classmethod
    def _item_from_dict(cls, item_dict: Dict[str, Any], generic_okay: bool = False, relaxed_validation: bool = False):
        generic_item_class: GenericType
        if relaxed_validation:
            registry = cls._RELAXED_TYPE_REGISTRY
            generic_item_class = cls._GENERIC_ITEM_CLASS_RELAXED_VALIDATION
        else:
            registry = cls._TYPE_REGISTRY
            generic_item_class = cls._GENERIC_ITEM_CLASS

        item_type = item_dict["category"]
        item_cls: Type[OPAbstractItemDescriptor]
        try:
            item_cls = registry[item_type]
        except KeyError as ke:
            if generic_okay:
                item_cls = generic_item_class
            else:
                raise OPUnknownItemTypeException(
                    f"Unknown item type {item_type}", item_dict=item_dict) from ke

        obj = item_cls(item_dict)

        return obj

    @classmethod
    def op_item(cls, item_json_or_dict: Union[str, Dict], generic_okay: bool = False, relaxed_validation: bool = False):
        """
        Factory methiod to instantiate an op item from JSON or a dictionary

        Parameters
        ----------
        item_json_or_dict : Union[str, Dict]
            JSON or dictionary representing an op item object
        relaxed_validation : bool, optional
            Whether relaxed validation should be enabled for this instance, by default False
            If true, the item class will be looked up from the relaxed validation registry

        Returns
        -------
        OPAbstractItem
            An instance of OPAbstractItem representing the 1Password item object

        Raises
        ------
        OPInvalidItemException
            If unserializing JSON fails, or the dictionary is otherwise invalid for an item object
        """
        try:
            item_dict = safe_unjson(item_json_or_dict)
        except JSONDecodeError as jdce:
            raise OPInvalidItemException(
                f"Failed to unserialize item JSON: {jdce}") from jdce
        obj = cls._item_from_dict(
            item_dict, generic_okay=generic_okay, relaxed_validation=relaxed_validation)
        return obj


def op_register_item_type(item_class):
    item_type = item_class.CATEGORY
    OPItemFactory.register_op_item_type(item_type, item_class)
    return item_class




import functools
import warnings
from typing import Any, Callable, Dict

"""
Adapted from:
https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/deprecation.py
"""


class deprecated:  # pragma: no cover

    def __init__(self, extra=""):
        self.extra = extra

    def __call__(self, obj):
        if isinstance(obj, type):
            return self._decorate_class(obj)
        else:
            return self._decorate_fun(obj)

    def _decorate_class(self, cls):
        msg = "Class {} is deprecated".format(cls.__name__)
        if self.extra:
            msg += ": {}".format(self.extra)

        init = cls.__init__

        def wrapped(_self, *args, **kwargs):
            if not getattr(_self, "_skip_drecation_warn", False):
                warnings.warn(msg, category=FutureWarning)
            return init(_self, *args, **kwargs)
        cls.__init__ = wrapped

        wrapped.__name__ = '__init__'
        wrapped.__doc__ = self._update_doc(init.__doc__)

        # mypy doesn't like wrapped.deprecated_original = init
        setattr(wrapped, "deprecated_original", init)

        return cls

    def _decorate_fun(self, fun):
        """Decorate function fun"""

        msg = "Function %s is deprecated" % fun.__name__
        if self.extra:
            msg += "; %s" % self.extra

        @functools.wraps(fun)
        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=FutureWarning)
            return fun(*args, **kwargs)

        wrapped.__doc__ = self._update_doc(wrapped.__doc__)

        return wrapped

    def _update_doc(self, olddoc):
        newdoc = "DEPRECATED"
        if self.extra:
            newdoc = "%s: %s" % (newdoc, self.extra)
        if olddoc:
            newdoc = "%s\n\n    %s" % (newdoc, olddoc)
        return newdoc


def deprecated_kwargs(**kwarg_aliases: str) -> Callable:
    """Decorator for deprecated function and method arguments.

    Use as follows:

    @deprecated_kwargs(old_arg='new_arg')
    def myfunc(new_arg):
        ...
    https://stackoverflow.com/a/49802489/17391340
    """

    def deco(f: Callable):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            _rename_kwargs(f.__name__, kwargs, kwarg_aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def _rename_kwargs(func_name: str, kwargs: Dict[str, Any], kwarg_aliases: Dict[str, str]):  # pragma: no cover
    """Helper function for deprecating function arguments."""
    for old_kwarg, new_kwarg in kwarg_aliases.items():
        if old_kwarg in kwargs:
            if new_kwarg in kwargs:
                raise TypeError(
                    f"{func_name} received both {old_kwarg} and {new_kwarg} as arguments!"
                    f" {old_kwarg} is deprecated, use {new_kwarg} instead."
                )
            warnings.warn(
                message=(
                    f"`{old_kwarg}` is deprecated as an argument to `{func_name}`; use"
                    f" `{new_kwarg}` instead."
                ),
                category=FutureWarning,
                stacklevel=3,
            )
            kwargs[new_kwarg] = kwargs.pop(old_kwarg)



import datetime
from typing import List, Optional

from ..._abc_meta import ABCMetaDict, enforcedmethod
from ..._datetime import fromisoformat_z
from ...json import safe_unjson
from ...op_objects import OPVaultDescriptor


class OPAbstractItemDescriptor(dict, metaclass=ABCMetaDict):
    ITEM_CATEGORY = None

    @enforcedmethod
    def __init__(self, item_dict_or_json):
        item_dict = safe_unjson(item_dict_or_json)
        super().__init__(item_dict)
        vault_dict = self.get("vault")
        if vault_dict:
            self._vault = OPVaultDescriptor(vault_dict)
        else:
            self._vault = None

    @property
    def vault(self) -> OPVaultDescriptor:
        return self._vault

    @property
    def unique_id(self) -> str:
        return self["id"]

    @property
    def title(self) -> str:
        title = self["title"]
        return title

    @property
    def tags(self) -> List[str]:
        return self.get("tags", [])

    @property
    def created_at(self) -> datetime.datetime:
        created = self["created_at"]
        created = fromisoformat_z(created)
        return created

    @property
    def updated_at(self) -> datetime.datetime:
        updated = self["updated_at"]
        updated = fromisoformat_z(updated)
        return updated

    @property
    def last_edited_by(self) -> str:
        return self["last_edited_by"]

    @property
    def vault_id(self) -> str:
        return self.vault.unique_id

    @property
    def state(self) -> Optional[str]:
        return self.get("state")

    @property
    def archived(self) -> bool:
        return self.state == "ARCHIVED"

    @property
    def favorite(self) -> bool:
        fav = self.get("favorite", False)
        return fav

    @property
    def version(self) -> int:
        return self["version"]

    @property
    def category(self) -> str:
        return self["category"]



from typing import Any, Dict, Type

from .fields_sections.item_field_base import OPItemField


class OPItemFieldFactory:
    _TYPE_REGISTRY: Dict[str, Type[OPItemField]] = {}

    @classmethod
    def register_op_field_type(cls, item_class):
        cls._TYPE_REGISTRY[item_class.FIELD_TYPE] = item_class

    @classmethod
    def field_type_lookup(cls, field_dict: Dict[str, Any]):
        item_type = field_dict["type"]
        field_class = OPItemField
        try:
            field_class = cls._TYPE_REGISTRY[item_type]
        except KeyError:
            pass
        return field_class

    @classmethod
    def item_field(cls, field_dict: Dict[str, Any], *args):
        item_cls = cls.field_type_lookup(field_dict)

        obj = item_cls(field_dict)
        return obj


def op_register_item_field_type(item_class):
    OPItemFieldFactory.register_op_field_type(item_class)
    return item_class



import sys
from argparse import ArgumentParser

from ._op_cli_config import (
    OPCLIAccountConfig,
    OPCLIConfig,
    OPConfigNotFoundException
)


def opc_parse_args():
    parser = ArgumentParser()
    parser.add_argument("--print-account-key",
                        help="Print just the account key instead of the whole config", action='store_true')
    parser.add_argument("--config-path", help="Path to 'op' configuration")
    parser.add_argument("--shorthand", help="Account shorthand to look up")
    parser.add_argument(
        "--raw", help="Write output suitable for piping to another process", action='store_true')

    parsed = parser.parse_args()
    return parsed


def print_config(acct_conf: OPCLIAccountConfig):
    print(f"shorthand: {acct_conf.shorthand}")
    print(f"URL: {acct_conf.url}")
    print(f"email: {acct_conf.email}")
    print(f"Account key: {acct_conf['accountKey']}")
    print(f"Account ID: {acct_conf.account_uuid}")
    print(f"User ID: {acct_conf.user_uuid}")


def main():
    options = opc_parse_args()
    config_path = options.config_path
    shorthand = options.shorthand
    try:
        config = OPCLIConfig(configpath=config_path)
        acct_conf = config.get_config(account_id=shorthand)
    except OPConfigNotFoundException as e:
        print(f"Unable to look up op config: {e}", file=sys.stderr)
        exit(1)

    if not options.raw:
        print(f"Config path: {config.configpath}")
    if options.print_account_key:
        if options.raw:
            sys.stdout.write(acct_conf["accountKey"])
            sys.stdout.flush()
        else:
            print(f"Account key: {acct_conf['accountKey']}")
    else:
        print_config(acct_conf)



import shlex
from typing import List, Optional, Union

from ._svc_account import OPSvcAcctSupportCode, OPSvcAcctSupportRegistry
from .op_items._new_item import OPNewItemMixin
from .op_items.password_recipe import OPPasswordRecipe


class _OPArgv(list):
    """
    This is essentially an 'argv' list with some additional methods, such as class methods
    for generating proper argument lists for specific 1Password operations.

    The primary purpose of this class is to facilitate the 'mock-op' project's automated response generation,
    as it allows the preciese set of command line arguments to be captured for later playback.
    """

    def __init__(self,
                 op_exe: str,
                 command: str,
                 args: List[str],
                 subcommands: Optional[Union[str, List[str]]] = None,
                 global_args: List[str] = [],
                 encoding="utf-8"):
        # TODO: Refactor this
        # constructor is getting too many specialized kwargs tied to
        # specific commands/subcommands
        # maybe instead of an "args" array plus a bunch of named kwargs,
        # send in a dict that gets passed through tree of argv building logic?

        argv = [op_exe]
        if not global_args:
            global_args = []
        else:
            global_args = list(global_args)

        if encoding.lower() != "utf-8":  # pragma: no coverage
            global_args.extend(["--encoding", encoding])

        for arg in global_args:
            argv.append(arg)
        if command:
            argv.append(command)
        self.command = command
        self.subcommands = None

        # whatever flags the command or subcommand take, plust global flags
        self.args_to_command = args

        if subcommands:
            if isinstance(subcommands, str):
                subcommands = [subcommands]
            else:
                subcommands = list(subcommands)  # pragma: no cover
            self.subcommands = subcommands
            argv.extend(subcommands)
        argv.extend(args)
        super().__init__(argv)

    def cmd_str(self):
        """
        return a shell-escaped command string from this argv
        """
        cmd_str = shlex.join(self)
        return cmd_str

    def svc_account_supported(self) -> OPSvcAcctSupportCode:
        # OPSvcAcctSupportRegistry is a singleton
        # so this is fine
        reg = OPSvcAcctSupportRegistry()
        supported = reg.command_supported(self)
        return supported

    @classmethod
    def item_generic_argv(cls,
                          op_exe: str,
                          subcommands: Optional[Union[str, List[str]]],
                          sub_cmd_args: Optional[List[str]] = None):
        args = []
        global_args = ["--format", "json"]
        if sub_cmd_args:
            args.extend(sub_cmd_args)
        argv = cls(op_exe,
                   "item",
                   args,
                   subcommands=subcommands,
                   global_args=global_args)
        return argv

    @classmethod
    def item_get_argv(cls,
                      op_exe,
                      item_name_or_id,
                      vault=None,
                      fields=None,
                      include_archive=False):
        sub_cmd_args = [item_name_or_id]
        if vault:
            sub_cmd_args.extend(["--vault", vault])

        if fields:
            sub_cmd_args.extend(["--fields", fields])
        if include_archive:
            sub_cmd_args.append("--include-archive")
        argv = cls.item_generic_argv(op_exe, "get", sub_cmd_args)
        return argv

    @classmethod
    def item_get_totp_argv(cls, op_exe, item_name_or_id, vault=None):
        field_arg = "type=otp"

        argv = cls.item_get_argv(
            op_exe, item_name_or_id, vault=vault, fields=field_arg)
        return argv

    @classmethod
    def document_generic_argv(cls,
                              op_exe: str,
                              subcommands: Optional[Union[str, List[str]]],
                              sub_cmd_args: Optional[List[str]]):
        args = []
        global_args = ["--format", "json"]
        if sub_cmd_args:
            args.extend(sub_cmd_args)
        argv = cls(op_exe,
                   "document",
                   args,
                   subcommands=subcommands,
                   global_args=global_args)
        return argv

    @classmethod
    def document_get_argv(cls, op_exe, document_name_or_id, vault=None, include_archive=False):

        sub_cmd_args = [document_name_or_id]
        if vault:
            sub_cmd_args.extend(["--vault", vault])
        if include_archive:
            sub_cmd_args.append("--include-archive")
        argv = cls.document_generic_argv(op_exe, "get", sub_cmd_args)
        return argv

    @classmethod
    def vault_generic_argv(cls,
                           op_exe: str,
                           subcommands: Optional[Union[str, List[str]]],
                           sub_cmd_args: Optional[List[str]]):
        args = []
        global_args = ["--format", "json"]
        if sub_cmd_args:
            args.extend(sub_cmd_args)
        argv = cls(op_exe, "vault", args, subcommands=subcommands,
                   global_args=global_args)
        return argv

    @classmethod
    def vault_get_argv(cls, op_exe, vault_name_or_id):
        sub_cmd_args = [vault_name_or_id]
        argv = cls.vault_generic_argv(op_exe, "get", sub_cmd_args)
        return argv

    @classmethod
    def vault_list_argv(cls, op_exe, group_name_or_id=None, user_name_or_id=None):
        sub_cmd_args = []
        if group_name_or_id:
            sub_cmd_args.extend(["--group", group_name_or_id])
        if user_name_or_id:
            sub_cmd_args.extend(["--user", user_name_or_id])
        argv = cls.vault_generic_argv(op_exe, "list", sub_cmd_args)
        return argv

    @classmethod
    def user_generic_argv(cls,
                          op_exe: str,
                          subcommands: Optional[Union[str, List[str]]],
                          sub_cmd_args: Optional[List[str]]):
        args = []
        global_args = ["--format", "json"]
        if sub_cmd_args:
            args.extend(sub_cmd_args)
        argv = cls(op_exe, "user", args, subcommands=subcommands,
                   global_args=global_args)
        return argv

    @classmethod
    def user_get_argv(cls, op_exe, user_name_or_id):
        sub_cmd_args = [user_name_or_id]
        argv = cls.user_generic_argv(op_exe, "get", sub_cmd_args)
        return argv

    @classmethod
    def user_list_argv(cls, op_exe, group_name_or_id=None, vault=None):
        sub_cmd_args = []
        if group_name_or_id:
            sub_cmd_args.extend(["--group", group_name_or_id])
        if vault:
            sub_cmd_args.extend(["--vault", vault])
        argv = cls.user_generic_argv(op_exe, "list", sub_cmd_args)
        return argv

    @classmethod
    def group_generic_argv(cls,
                           op_exe: str,
                           subcommands: Optional[Union[str, List[str]]],
                           sub_cmd_args: Optional[List[str]]):
        args = []
        global_args = ["--format", "json"]
        if sub_cmd_args:
            args.extend(sub_cmd_args)
        argv = cls(op_exe, "group", args, subcommands=subcommands,
                   global_args=global_args)
        return argv

    @classmethod
    def group_get_argv(cls, op_exe, user_name_or_id):
        sub_cmd_args = [user_name_or_id]
        argv = cls.group_generic_argv(op_exe, "get", sub_cmd_args)
        return argv

    @classmethod
    def group_list_argv(cls, op_exe, user_name_or_id=None, vault=None):
        sub_cmd_args = []
        if user_name_or_id:
            sub_cmd_args.extend(["--user", user_name_or_id])
        if vault:
            sub_cmd_args.extend(["--vault", vault])
        argv = cls.group_generic_argv(op_exe, "list", sub_cmd_args)
        return argv

    @classmethod
    def normal_signin_argv(cls, op_exe, account=None):
        global_args = []
        if account:
            global_args = ["--account", account]
        argv = ["--raw"]
        return cls(op_exe, "signin", argv, global_args=global_args)

    @classmethod
    def item_template_list_argv(cls, op_exe):  # pragma: no cover
        # subcommands may be a string or list
        # so no need to put the sub-subcommand
        # into an arg list
        sub_commands = ["template", "list"]
        sub_cmd_args: List[str] = []

        argv_obj = cls.item_generic_argv(op_exe, sub_commands, sub_cmd_args)
        return argv_obj

    @classmethod
    def whoami_argv(cls, op_exe, account=None):
        args: List[str] = []
        global_args = []
        if account:
            global_args = ["--account", account]
        global_args.extend(["--format", "json"])
        argv_obj = cls(op_exe, "whoami", args, global_args=global_args)
        return argv_obj

    @classmethod
    def cli_version_argv(cls, op_exe):
        args: List[str] = []
        global_args = ["--version"]
        argv_obj = cls(op_exe, None, args, global_args=global_args)
        return argv_obj

    @classmethod
    def signout_argv(cls, op_exe, account_shorthand: str, session: str, forget=False, uses_bio=False):  # pragma: no cover
        global_args = []
        if not uses_bio:
            global_args = ["--account",
                           account_shorthand, "--session", session]
        signout_args = []
        if forget:
            signout_args.append("--forget")
        argv = cls(op_exe, "signout", signout_args, global_args=global_args)
        return argv

    @classmethod
    def item_list_argv(cls, op_exe, categories=[], include_archive=False, tags=[], vault=None):
        item_list_args = []
        if categories:
            categories_arg = ",".join(categories)
            item_list_args.extend(["--categories", categories_arg])
        if include_archive:
            item_list_args.append("--include-archive")
        if tags:
            tags_args = ",".join(tags)
            item_list_args.extend(["--tags", tags_args])
        if vault:
            item_list_args.extend(["--vault", vault])

        argv = cls.item_generic_argv(op_exe, "list", item_list_args)
        return argv

    @classmethod
    def account_generic_argv(cls,
                             op_exe: str,
                             subcommands: Optional[Union[str, List[str]]],
                             sub_cmd_args: Optional[List[str]] = None,
                             encoding: str = "utf-8"):
        args = []
        cmd = "account"
        global_args = ["--format", "json"]
        if sub_cmd_args:  # pragma: no coverage
            args.extend(sub_cmd_args)
        argv = cls(op_exe, cmd, args, subcommands=subcommands,
                   global_args=global_args, encoding=encoding)
        return argv

    @classmethod
    def account_list_argv(cls, op_exe, encoding="utf-8"):
        subcmd = "list"
        argv = cls.account_generic_argv(
            op_exe, subcmd, encoding=encoding)
        return argv

    @classmethod
    def account_forget_argv(cls, op_exe, account):  # pragma: no coverage
        subcmd = "forget"
        sub_cmd_args = [account]
        argv = cls.account_generic_argv(
            op_exe, subcmd, sub_cmd_args=sub_cmd_args)

        return argv

    @classmethod
    def item_create_argv(cls,
                         op_exe,
                         item: OPNewItemMixin,
                         password_recipe: Optional[OPPasswordRecipe] = None,
                         vault: Optional[str] = None,
                         encoding="utf-8"):
        """
        op item create --template ./new_item.json --vault "Test Data" --generate-password=20,letters,digits --dry-run --format json
        """
        template_filename = item.secure_tempfile(
            encoding=encoding)

        item_create_args = ["--template", template_filename]

        if password_recipe:
            # '--password-recipe' requires an '=' unlike other option/argument pairs
            item_create_args.append(f"--generate-password={password_recipe}")
        if vault:
            item_create_args.extend(["--vault", vault])
        argv = cls.item_generic_argv(op_exe, "create")
        return argv

    @classmethod
    def item_delete_argv(cls,
                         op_exe: str,
                         item_name_or_id: str,
                         vault: Optional[str] = None,
                         archive: bool = False):
        sub_cmd_args = [item_name_or_id]
        if archive:
            sub_cmd_args.append("--archive")
        if vault:
            sub_cmd_args.extend(["--vault", vault])
        delete_argv = cls.item_generic_argv(op_exe, "delete", sub_cmd_args)

        return delete_argv

    @classmethod
    def document_delete_argv(cls,
                             op_exe: str,
                             document_name_or_id: str,
                             vault: Optional[str] = None,
                             archive: bool = False):
        sub_cmd_args = [document_name_or_id]
        if archive:
            sub_cmd_args.append("--archive")
        if vault:
            sub_cmd_args.extend(["--vault", vault])
        delete_argv = cls.document_generic_argv(op_exe, "delete", sub_cmd_args)

        return delete_argv



import base64
import re
import secrets


class OPUniqueIdentifierBase32:
    BYTE_COUNT = 16

    def __init__(self, uppercase=False):
        # use secrets.token_bytes() rather than random.randbytes()
        # mainly because randbytes() isn't available < python 3.9
        # It's also  better for cryptographic secrets, but
        # that shouldn't matter here
        self._random_bytes = secrets.token_bytes(self.BYTE_COUNT)
        self._upper = uppercase

    def __str__(self):
        unique_id_bytes = self._random_bytes
        # more mypy hax
        # can't reuse b32unique_id, so need a separate bytes var
        b32unique_id_bytes = base64.b32encode(unique_id_bytes)
        b32unique_id = b32unique_id_bytes.decode("utf-8")
        b32unique_id = b32unique_id.rstrip('=')
        if not self._upper:
            b32unique_id = b32unique_id.lower()
        return b32unique_id


class OPUniqueIdentifierHex(OPUniqueIdentifierBase32):  # pragma: no coverage
    PREFIX = ""

    def __init__(self, uppercase=True):
        super().__init__(uppercase=uppercase)

    def __str__(self):
        unique_id_bytes = self._random_bytes
        unique_id = unique_id_bytes.hex()
        if self._upper:
            unique_id = unique_id.upper()
        unique_id = f"{self.PREFIX}{unique_id}"
        return unique_id


class OPUniqueSectionIdentifier(OPUniqueIdentifierHex):
    PREFIX = "Section_"


def is_uuid(id_str: str):
    b32_regex = r'^[A-Z0-9]{26}$'
    b32_lowercase_regex = r'^[a-z0-9]{26}$'
    hex_bytes_regex = r'[a-zA-Z0-9]{32}$'
    section_id_regex = r'Section_[a-zA-Z0-9]{32}$'

    match = False
    for regex in [b32_regex, b32_lowercase_regex, hex_bytes_regex, section_id_regex]:
        if re.match(regex, id_str):
            match = True
            break

    return match



from pyonepassword import OP


def _about_string():
    from pyonepassword.__about__ import __summary__, __title__, __version__
    about_str = f"{__title__.upper()}: {__summary__}. Version {__version__}"
    return about_str


def test_pyop_about_01():
    about_str = _about_string()
    assert OP.about() == about_str


def test_pyop_version_01():
    from pyonepassword.__about__ import __version__
    assert OP.version() == __version__



__title__ = "pyonepassword"
__version__ = "3.11.0"
__summary__ = "A python API to query a 1Password account using the 'op' command-line tool"

"""
See PEP 440 for version scheme
https://www.python.org/dev/peps/pep-0440/#examples-of-compliant-version-schemes
Examples:

FINAL
0.9
0.9.1
0.9.2
...
0.9.10
0.9.11
1.0
1.0.1
1.1
2.0
2.0.1
...

PRE_RELEASES
X.Y.dev0 # development release 0
x.y.dev1 # development release 1
X.YaN    # Alpha release
X.YbN    # Beta release
X.YrcN   # Release Candidate
X.Y      # Final release
"""



from abc import ABCMeta


def enforcedmethod(func):
    """
    decorator similar to @abstractmethod
    to enforce that a method be overridden
    """
    func.__enforcedmethod__ = True
    return func


# TODO: we really should be subclassing 'type' here, but
# mypy incorrectly complains about metaclass conflict
# https://github.com/python/mypy/issues/14182#issuecomment-1407216483
class ABCMetaDict(ABCMeta):
    """
    Metaclass that enables a class descending from a concrete class
    to still enforce the requirement that one or more methods be overridden
    E.g.:

    class _OPItemBase(dict):

        @enforcedmethod
        def __init__(self, *args, **kwargs):
            # you must override this method, and then call this method via super()
    """

    def __call__(cls, *args, **kwargs):
        enforced = set()
        unenforced = set()

        # get two sets of attributes: enforced flag is set and enforced flag is not set
        for name, value in cls.__dict__.items():
            if getattr(value, "__enforcedmethod__", False):
                enforced.add(name)
            else:
                unenforced.add(name)

        for base in cls.__mro__:
            if base in [dict, object, cls]:
                # if the class is dict, object or this class, skip it
                # that way methods in those classes don't count as being imlemented
                continue

            for name, value in base.__dict__.items():
                # if the item is flagged as enforced and isn't already in the
                # unenforced set (i.e., there's an implementation somewhere)
                # then added to the enforced set
                if getattr(value, "__enforcedmethod__", False):
                    if name not in unenforced:
                        enforced.add(name)
                else:
                    # this item isn't flagged as enforced, so add it to the unenforced list
                    unenforced.add(name)

        if enforced:
            raise TypeError("Can't instantiate abstract class {} "
                            "with enforced methods {}".format(
                                cls.__name__, ', '.join(enforced)))
        else:
            return super(ABCMetaDict, cls).__call__(*args, **kwargs)



from ..op_items._item_type_registry import op_register_item_type

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = [
    "op_register_item_type"
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyonepassword import OP

import pytest

from pyonepassword.api.exceptions import (
    OPDocumentGetException,
    OPInvalidDocumentException
)
from pyonepassword.api.object_types import OPDocumentFile, OPDocumentItem

from .test_support.util import digest

pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_document_get01(signed_in_op: OP, expected_document_data):
    item_name = "Example Login 2 - 1200px-SpongeBob_SquarePants_character.svg.png.webp"
    vault = "Test Data"
    expected = expected_document_data.data_for_document(item_name)
    filename, data = signed_in_op.document_get(item_name, vault=vault)
    sha256_digest = digest(data)
    assert filename == expected.filename
    assert len(data) == expected.size
    assert sha256_digest == expected.digest


@pytest.mark.parametrize("invalid_document,vault",
                         [("Invalid Document", None),
                          ("Error Success", "Test Data"),
                          ("Example Attached File 2", None)])
def test_document_get02(signed_in_op: OP, expected_document_data, invalid_document, vault):
    exception_class = OPDocumentGetException
    expected = expected_document_data.data_for_document(invalid_document)
    try:
        signed_in_op.document_get(invalid_document, vault=vault)
        assert False, f"We should have caught {exception_class.__name__}"
    except exception_class as e:
        assert expected.returncode == e.returncode


def test_document_get03(signed_in_op: OP, expected_document_data):
    item_name = "Example Login 2 - 1200px-SpongeBob_SquarePants_character.svg.png.webp"
    vault = "Test Data"
    expected: OPDocumentItem = expected_document_data.data_for_document(
        item_name)
    document_item = signed_in_op.item_get(item_name, vault=vault)
    assert isinstance(document_item, OPDocumentItem)
    document_files = document_item.files
    file_obj = document_files[0]
    assert isinstance(file_obj, OPDocumentFile)
    assert file_obj.size == expected.size
    assert file_obj.content_path == expected.content_path
    assert file_obj.file_id == expected.file_id


def test_document_get_wrong_item_type_01(signed_in_op: OP):
    with pytest.raises(OPInvalidDocumentException):
        signed_in_op.document_get("Not A Document")



# needed so we can pass getattr() logic
# through to module for deprecation warnings
from .. import py_op_exceptions as __py_op_exceptions
from .._svc_account import OPSvcAcctCommandNotSupportedException
from ..op_items._item_type_registry import OPUnknownItemTypeException
from ..op_items.fields_sections._new_fields import OPNewTOTPUriException
from ..op_items.fields_sections.item_section import (
    OPItemFieldCollisionException,
    OPSectionCollisionException
)
from ..op_items.item_types._item_base import (
    OPFieldNotFoundException,
    OPSectionNotFoundException
)
from ..op_items.item_types.login import OPNewLoginItemURLException
from ..op_objects import (
    OPInvalidGroupException,
    OPInvalidGroupListException,
    OPInvalidObjectException,
    OPInvalidUserException,
    OPInvalidUserListException,
    OPInvalidVaultException,
    OPInvalidVaultListException
)
from ..py_op_exceptions import (
    OPAuthenticationException,
    OPCLIPanicException,
    OPCmdFailedException,
    OPCmdMalformedSvcAcctTokenException,
    OPConfigNotFoundException,
    OPDocumentDeleteException,
    OPDocumentGetException,
    OPForgetException,
    OPGroupGetException,
    OPGroupListException,
    OPInvalidDocumentException,
    OPInvalidItemException,
    OPItemDeleteException,
    OPItemDeleteMultipleException,
    OPItemGetException,
    OPItemListException,
    OPNotFoundException,
    OPRevokedSvcAcctTokenException,
    OPSigninException,
    OPSignoutException,
    OPUnknownAccountException,
    OPUserGetException,
    OPUserListException,
    OPVaultGetException,
    OPVaultListException
)

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = ["OPAuthenticationException",
           "OPCLIPanicException",
           "OPCmdFailedException",
           "OPCmdMalformedSvcAcctTokenException",
           "OPConfigNotFoundException",
           "OPDocumentDeleteException",
           "OPDocumentGetException",
           "OPFieldNotFoundException",
           "OPForgetException",
           "OPGroupGetException",
           "OPGroupListException",
           "OPInvalidDocumentException",
           "OPInvalidGroupException",
           "OPInvalidGroupListException",
           "OPInvalidItemException",
           "OPInvalidObjectException",
           "OPInvalidUserException",
           "OPInvalidUserListException",
           "OPInvalidVaultException",
           "OPInvalidVaultListException",
           "OPItemDeleteException",
           "OPItemDeleteMultipleException",
           "OPItemFieldCollisionException",
           "OPItemGetException",
           "OPItemListException",
           "OPNewLoginItemURLException",
           "OPNewTOTPUriException",
           "OPNotFoundException",
           "OPSectionCollisionException",
           "OPSectionNotFoundException",
           "OPRevokedSvcAcctTokenException",
           "OPSigninException",
           "OPSignoutException",
           "OPSvcAcctCommandNotSupportedException",
           "OPUnknownAccountException",
           "OPUnknownItemTypeException",
           "OPUserGetException",
           "OPUserListException",
           "OPVaultGetException",
           "OPVaultListException"]


def __getattr__(name: str):
    return getattr(__py_op_exceptions, name)

    # if name in _deprecated_exceptions:
    #     _deprecated_name = f"_{name}"
    #     alternate = _deprecated_exceptions[name]
    #     warnings.warn(
    #         f"Exception class {name} is deprecated. Use {alternate}", category=FutureWarning)
    #     return globals()[_deprecated_name]

    # raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import json
import os
import pathlib
from json.decoder import JSONDecodeError
from typing import List, Optional

from .py_op_exceptions import OPConfigNotFoundException


class OPCLIAccountConfig(dict):

    def __init__(self, account_dict):
        super().__init__(account_dict)

    @property
    def shorthand(self) -> str:
        return self["shorthand"]

    @property
    def account_uuid(self) -> str:
        return self["accountUUID"]

    @property
    def url(self) -> str:
        return self["url"]

    @property
    def email(self) -> str:
        return self["email"]

    @property
    def user_uuid(self) -> str:
        return self["userUUID"]


class OPCLIConfig(dict):
    OP_CONFIG_PATHS = [
        pathlib.Path(".config", "op", "config"),
        pathlib.Path(".op", "config")
    ]

    def __init__(self, configpath=None):
        super().__init__()
        if configpath is None:
            configpath = self._get_config_path()
        self.configpath = configpath
        if configpath is None:
            raise OPConfigNotFoundException("No op configuration found")
        self._configpath = configpath
        try:
            config_json = open(configpath, "r").read()
        except FileNotFoundError as e:
            raise OPConfigNotFoundException(
                "op config not found at path: {}".format(configpath)) from e
        except PermissionError as e:
            raise OPConfigNotFoundException(
                "Permission denied accessing op config at path: {}".format(configpath)) from e

        try:
            config = json.loads(config_json)
            self.update(config)
        except JSONDecodeError as e:
            raise OPConfigNotFoundException(
                "Unable to json decode config at path: {}".format(configpath)) from e

        accounts = self._initialize_account_objects()
        self["accounts"] = accounts

        account_map = {}
        for account in accounts:
            account_map[account.shorthand] = account
        self.account_map = account_map

    def _get_config_path(self) -> pathlib.Path:
        configpath: pathlib.Path = None
        config_home = None
        try:
            config_home = pathlib.Path(os.environ['XDG_CONFIG_HOME'])

        except KeyError:
            config_home = pathlib.Path.home()

        for subpath in self.OP_CONFIG_PATHS:
            _configpath = pathlib.Path(config_home, subpath)
            if os.path.exists(_configpath):
                configpath = _configpath
                break

        return configpath

    def _initialize_account_objects(self) -> List[OPCLIAccountConfig]:
        account_list = self.accounts
        account_objects = []
        acct: OPCLIAccountConfig
        for account_dict in account_list:
            acct = OPCLIAccountConfig(account_dict)
            account_objects.append(acct)

        return account_objects

    @property
    def accounts(self) -> List[OPCLIAccountConfig]:
        account_list = self.get("accounts")
        if account_list is None:
            account_list = []
        return account_list

    @property
    def latest_signin(self) -> Optional[str]:
        return self.get("latest_signin")

    @property
    def latest_signin_uuid(self) -> Optional[str]:
        latest_uuid = None
        latest = self.latest_signin
        if latest:
            latest_uuid = self.uuid_for_account(latest)
        return latest_uuid

    def get_config(self, account_id=None) -> OPCLIAccountConfig:
        if account_id is None:
            account_id = self.get("latest_signin")
        if not account_id:  # if shorthand is None or empty string
            raise OPConfigNotFoundException(
                "No account identifier provided, no sign-ins found.")

        config = self.account_map.get(account_id)
        if not config:
            for account in self.accounts:
                if account_id in [account.account_uuid, account.user_uuid, account.shorthand, account.email, account.url]:
                    config = account
                    break

        if config is None:
            raise OPConfigNotFoundException(
                f"No config found for account identifier '{account_id}'")
        return config

    def uuid_for_account(self, account_identifier) -> str:
        config = self.get_config(account_id=account_identifier)
        uuid = config.user_uuid
        return uuid



class OPInvalidPasswordRecipeException(Exception):
    pass


class OPPasswordRecipe:
    def __init__(self, length: int = 20, letters=True, digits=True, symbols=True):
        recipe_parts = [str(length)]

        if True not in [letters, digits, symbols]:
            raise OPInvalidPasswordRecipeException(
                "Letters, digits, & symbols all disabled. One or more must be enabled")

        if letters:
            recipe_parts.append("letters")
        if digits:
            recipe_parts.append("digits")
        if symbols:
            recipe_parts.append("symbols")

        self.recipe = recipe_parts

    def __str__(self) -> str:
        recipe_str = ",".join(self.recipe)
        return recipe_str


LETTERS_DIGITS_SYMBOLS_20 = OPPasswordRecipe()
LETTERS_DIGITS_25 = OPPasswordRecipe(length=25, symbols=False)



# __future__.annotations, and typing.TYPE_CHECKING
# enable anything imported for type hinting to disappear at run time
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from pyonepassword.api.exceptions import OPDocumentDeleteException

# make imports for type-hinting disappear at run-time to avoid
# circular imports.
# this also reduced exercising tested code simply by importing
if TYPE_CHECKING:
    from pyonepassword import OP

    from .fixtures.expected_document_data import ExpectedDocumentData

from .test_support.util import digest

# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_document_delete_01(signed_in_op: OP, expected_document_data: ExpectedDocumentData):
    """
    Test deleting a document based on its non-unique title
    """
    document_name = "delete this document"
    vault = "Test Data"
    expected = expected_document_data.data_for_document(document_name)
    expected_document_id = expected.item_id
    result = signed_in_op.document_delete(document_name, vault=vault)
    assert result == expected_document_id


def test_document_delete_02(signed_in_op: OP, expected_document_data: ExpectedDocumentData):
    """
    Test deleting and archiving a document based on its non-unique title
    """
    document_name = "delete and archive this document"
    vault = "Test Data"
    expected = expected_document_data.data_for_document(document_name)
    expected_item_id = expected.item_id
    result = signed_in_op.document_delete(
        document_name, vault=vault, archive=True)
    assert result == expected_item_id


def test_document_delete_03(signed_in_op: OP, expected_document_data: ExpectedDocumentData):
    """
    Test:
      - deleting and archiving a document based on its non-unique title
      - fetching the item with include_archive=True
    """
    document_name = "delete and archive this document"
    vault = "Test Data"
    expected = expected_document_data.data_for_document(document_name)

    signed_in_op.document_delete(
        document_name, vault=vault, archive=True)
    filename, data = signed_in_op.document_get(
        document_name, vault=vault, include_archive=True)
    sha256_digest = digest(data)

    assert filename == expected.filename
    assert len(data) == expected.size
    assert sha256_digest == expected.digest


def test_document_delete_non_existent_01(signed_in_op: OP):
    """
    Test deleting a non-existent document
    """
    document_name = "non-existent-item"
    vault = "Test Data"

    with pytest.raises(OPDocumentDeleteException):
        signed_in_op.document_delete(document_name, vault=vault)


def test_document_delete_non_existent_02(signed_in_op: OP):
    """
    Test deleting a non-existent item, bypassing item_get()
    """
    document_name = "non-existent-item"
    vault = "Test Data"

    with pytest.raises(OPDocumentDeleteException):
        # OP.document_delete() calls item_get() first
        # which will fail on non-existent items
        # so in order to test delete operatation failing, we
        # need to call private ._document_delete() interface
        signed_in_op._document_delete(document_name, vault=vault)


def test_document_delete_duplicate_01(signed_in_op: OP):
    """
    Test:
      - deleting a document based on its non-unique title
      - when there are two or more documents sharing the same title
    """
    document_name = "delete this duplicate document"
    vault = "Test Data"

    with pytest.raises(OPDocumentDeleteException):
        signed_in_op.document_delete(document_name, vault=vault)
